/* -----------------------------------------------------------------------------
*
*  npconsole.c
*
*  ANTz - realtime 3D data visualization tools for the real-world, based on NPE.
*
*  ANTz is hosted at http://openantz.com and NPE at http://neuralphysics.org
*
*  Written in 2010-2012 by Shane Saxon - contact@saxondigital.net
*
*  Please see main.c for a complete list of additional code contributors.
*
*  To the extent possible under law, the author(s) have dedicated all copyright 
*  and related and neighboring rights to this software to the public domain
*  worldwide. This software is distributed without any warranty.
*
*  Released under the CC0 license, which is GPL compatible.
*
*  You should have received a copy of the CC0 Public Domain Dedication along
*  with this software (license file named COPYING.txt). If not, see
*  http://creativecommons.org/publicdomain/zero/1.0/
*
* --------------------------------------------------------------------------- */

#include "npconsole.h"

#include "../npdata.h"

#ifdef NP_MSW_
	#include <freeglut.h>
#endif
#ifdef NP_OSX_
	#include <GLUT/glut.h>
	#include <OpenGL/gl.h>
#endif
#ifdef NP_LINUX_
	#include <GL/freeglut.h>
#endif

#include "../npctrl.h"


/* -----------------------------------------------------------------------------
* multiple consoles in ASCII, 2D and 3D object texture mapped with hybrid modes
* per console event handling mapped from multiple user input devices
* --------------------------------------------------------------------------- */

void npKeyEvent (int key, int keyEventType);

//------------------------------------------------------------------------------
void npInitConsole (void* dataRef)
{
	return;
}


//------------------------------------------------------------------------------
void npCloseConsole (void* dataRef)
{
	return;
}

void npConsoleClearPrompt(pNPconsole console, void* dataRef);
//------------------------------------------------------------------------------
void npConsoleClearPrompt(pNPconsole console, void* dataRef)
{
	npPostMsg(">", kNPmsgView, dataRef);

	//clear the input string
	console->inputStr[0] = '\0';
	console->inputIndex = 0;

	//set cursor position
	console->cursorLine = 0;
	console->cursorColumn = 1;			//zz debug, should be 2 and no col zero
	console->cursorShow = true;
}

//------------------------------------------------------------------------------
void npConsoleKeyEvent (int key, int keyEventType, void* dataRef)
{
	int i = 0;
	int topItem = 1;
	int bottomItem = 20;
	int lastPage = 0;
	int itemChosen = 0;
	int activeItem = 0;
	int lineStart = 0;

	char msg[256];

	pData data = (pData) dataRef;
	pNPconsole console = &data->io.gl.hud.console;


	switch (keyEventType)
	{
		//first process specific ASCII keys
		case kKeyDownASCII :
			switch (key)
			{
				case 8  : // kKeyCodeBackSpace
					if(console->inputIndex > 0)	//delete only if input exists
					{
						//perhaps rename inputIndex to inputCount
						//translate col row to cursorIndex or store separately? zz debug
						//if deleting in the middle of string then compact the gap
					//	for(i = console->cursorIndex; i < console->inputIndex; i++)
					//		console->inputStr[i] = console->inputStr[i+1];
						
						//replace the last char with null terminator
						console->inputStr[console->inputIndex] = '\0';
						console->inputIndex--;
						
						console->cursorColumn--;
					}
					npUpdateConsoleUserText(console, data);
					if (console->cursorColumn == 0)
						console->line[console->lineIndex][0] = '\0';
					break;

//				case 32 : //kKeyCodeSpace : //display next page if it exists
					// console->page++;
					// npUpdateConsoleMenu(console, data);
					// break;
				case 13 : //kKeyCodeReturn :
					//process command string
					itemChosen = npatoi(console->inputStr);
					if (itemChosen > 0 && itemChosen <=  console->menu->count)
					{	
						sprintf(msg,"item %s", console->menu->list[itemChosen]);
						npPostMsg(msg, kNPmsgView, data);

						//exit menu mode
						console->mode = kNPconsoleMessage;
						console->cursorShow = false;
						npPostMsg("Exit Menu", kNPmsgCtrl, data);			//send as kNPmsgCtrl
						npPostMsg("Console Mode: message", kNPmsgCtrl, data);

						console->pMenuCallback(itemChosen, dataRef);
					}
					else
					{						
						// npUpdateConsoleMenu(console, data);
						sprintf(msg,"Unknown Item - range is 1 to %d", console->menu->count);
						npPostMsg(msg, kNPmsgView, data);
 
						npConsoleClearPrompt(console, data);
					}
					break;
				case 27 : //kKeyCodeESC :
					//return to default console mode when done
					console->mode = kNPconsoleMessage;
					console->cursorShow = false;
					npPostMsg("Exit Menu", kNPmsgCtrl, data);				//send as kNPmsgCtrl
					npPostMsg("Console Mode: message", kNPmsgCtrl, data);
					free(console->menu->list);
					break;

				default :
					//add key string storage with cursor position update
					//	console->page = 0;							//zero to scroll current line
					//	console->cursorLine = 0;
					if(console->inputIndex > kNPconsoleInputMax)
						return;

					console->inputStr[console->inputIndex++] = (char)key;

					console->cursorColumn++;

					npUpdateConsoleUserText(console, data);
					break;
			}
			break;

		//process Special NON-ASCII keys
		case kKeyDownSpecial :			// non-ASCII Glut constants
			switch (key) 
			{
								
				case kKeyCodeRight : break;
				case kKeyCodeLeft :
	/*				console->cursorColumn--;
					if(console->cursorColumn < 0)
					{
						console->cursorColumn = console->charPerLine - 1;
						console->cursorLine = 
					}
	*/				break;
				case kKeyCodeUp :								// UP arrow
					console->page--;
					if (console->page < 0)	//rollover to last page
						console->page = console->menu->count / kNPmenuListMax;
					
	/*	
					//update user text prompt with newly selected item number
					sprintf(msg, "%d", console->selectIndex);
				//	npConsolePrompt(msgStr, console, dataRef)
		*/			
					npUpdateConsoleMenu(console, data);
					break;

				case kKeyCodeDown :	// DOWN arrow
					console->page++;						//calc last page
					if (console->page > (console->menu->count / kNPmenuListMax))
						console->page = 0;
/*					console->selectIndex++;
					if (console->selectIndex > lastItem)
					{
						if(console->page > lastPage)
						{
							console->page = 0;
							console->selectIndex = 1;
						}
					}
					//update user text prompt with newly selected item number
					sprintf(msg, "%d", console->selectIndex);
					npPostMsg(msg, kNPmsgView, data);
*/
					npUpdateConsoleMenu(console, data);
					break;
			}
			break;
		}
	return;
}


//------------------------------------------------------------------------------
void npUpdateConsole (void* dataRef)
{
	int cmdLineCount = 0;
	double halfSecTest = 0.0;

	pData data = (pData) dataRef;
	pNPconsole console = NULL;
	
	//currently supports only 1 console,  make console a node type		//zz debug 
//	console = &data->io.gl.hud.console;

	//zz debug, move this code chunk to npTime()
	//subtract the integer portion of the time to get the decimal portion
	halfSecTest = data->io.time - (int)data->io.time;
	//set the blink cursor state every half sec, one sec for full cycle
	//inverse cursor drawn on top of text (in GL) when blinkState is true
	if ( halfSecTest < 0.5 ) 
		data->io.blinkState = true;
	else
		data->io.blinkState = false;

									//call this from event handler instead //zz debug
//	npUpdateConsoleUserText(&data->io.gl.hud.console, data);	
}

//clear console screen
//------------------------------------------------------------------------------
void npConsoleCLS (pNPconsole console, void* dataRef)
{
	int i = 0;

	for (i=0; i < kNPconsoleLineMax; i++)
		console->line[i][0] = '\0';
//		npPostMsg("",kNPmsgView, dataRef);

	return;
}

//------------------------------------------------------------------------------
void npUpdateConsoleUserText(pNPconsole console, void* data)
{
	int lineStartChar = 0;
	int currentLine = 0;
	int lineCount = 0;
	char str[kNPconsoleCharPerLine + 1];

	//if console input is active, then update the cursor and user text
	if (console->mode != kNPconsoleMenu)
		return;

		//handle line wrap, post input string to advance to new line
	if (console->cursorColumn >= console->charPerLine)
	{
		if (console->inputIndex < (console->charPerLine - 1) )
		{
			console->cursorColumn = 0;
			npPostMsg(console->inputStr, kNPmsgView, data);
		}
		else
		{
			console->cursorColumn = 0;
		
			lineStartChar = ((console->inputIndex + 1) / console->charPerLine) 
					* console->charPerLine; 

			npPostMsg(&console->inputStr[lineStartChar], kNPmsgView, data);	
		}
	} 
	else if (console->cursorColumn < 0)
	{
		console->cursorColumn = console->charPerLine - 1;
		console->lineIndex--;
	}

	// strncpy(userText, console->inputStr[0], console->inputIndex);
	// sprintf(msg,">%d %d %s", console->inputIndex, strlen(console->inputStr), console->inputStr);

	//ascerts the string is terminated,//zz debug
	console->inputStr[console->inputIndex] = '\0';
	lineCount = (console->inputIndex / kNPconsoleCharPerLine);

	//format the active line
	//first line has one less character
	if (console->inputIndex < console->charPerLine)	
		sprintf(str,">%s", console->inputStr);
	else
	{
		currentLine = (kNPconsoleCharPerLine * lineCount) - 1;
		strncpy(str, &console->inputStr[currentLine], console->charPerLine);
	}
	//copy formatted string to the console line buffer
	if  (console->inputIndex == (console->charPerLine - 1))
	{
		strncpy(&console->line[console->lineIndex][0], str, console->charPerLine);
	}
	else
		strncpy(&console->line[console->lineIndex][0], str, console->charPerLine);

}

//------------------------------------------------------------------------------
void npUpdateConsoleMenu (pNPconsole console, void* dataRef)
{
	int i = 0;
	int firstItem = 0;		//first menu item on page
	int lastItem = 0;		//last item on page
	int pageItemCount = 0;
	char lineStr[256];

	pNPmenu menu = NULL;
	pData data = (pData) dataRef;

	if(!console)
	{
		npPostMsg("err 4284 - null console ptr, npConsoleMenu", kNPmsgErr, data);
		return;
	}

	menu = console->menu;

	npConsoleCLS(console, data);	//clear screen

	//display names and paths of indexed files from paths table
	//page at a time with index number next to each entry for user to punch in
	if (1)//menu->count > kNPmenuListMax)
	{
		firstItem = console->page * kNPmenuListMax + 1;

		//if last page then calculate number of items on last page
		if ((menu->count - firstItem) < kNPmenuListMax)
			pageItemCount = menu->count - firstItem;
		else
			pageItemCount = kNPmenuListMax;	//not the last page set to max

		lastItem = firstItem + pageItemCount;

		//build this out for mouse and cursor to highlight specific item...    zz debug
		//arrows as page up/down, alternate method is press 'Spacebar' for next
		sprintf(lineStr,"Showing %d to %d (of %d) - press up/down arrow keys for next page", 
				firstItem, lastItem, menu->count);
		npPostMsg (lineStr, kNPmsgView, dataRef);
	}
	else
	{
		firstItem = 1;
		pageItemCount = menu->count;
		lastItem = firstItem + pageItemCount;
	}

	menu->name = npNewStrcpy("DB Name", data);
	menu->details = npNewStrcpy("Lccation", data);

	//format the menubar
	npPostMsg ("+-----+------------------+------------------------------------------+", kNPmsgView, dataRef);
	sprintf (lineStr, "|   # | %16s | %40s |", menu->name, menu->details);
	npPostMsg (lineStr, kNPmsgView, dataRef);	
	npPostMsg ("+-----+------------------+------------------------------------------+", kNPmsgView, dataRef);

	for (i = firstItem; i <= lastItem; i++)
		npPostMsg (menu->list[i], kNPmsgView, dataRef);

	npPostMsg ("+-----+------------------+------------------------------------------+", kNPmsgView, dataRef);
	
	npPostMsg ("", kNPmsgView, dataRef);
	npPostMsg ("Enter item number:", kNPmsgView, dataRef);
		
	npConsoleClearPrompt (console, data);	//clear and set user prompt

//	npPostMsg (">", kNPmsgView, dataRef);
}



//displays a menu list passed in and registers callback function then returns
//selected item sent to callback function upon user action
//------------------------------------------------------------------------------
void npConsoleMenu (void (*pMenuCallback)(int menuItem, void* dataRef),
					pNPconsole console, void* dataRef)
{
	pNPmenu menu = NULL;
	pData data = (pData) dataRef;

	if(!console)
	{
		npPostMsg("err 4284 - null console ptr, npConsoleMenu", kNPmsgErr, data);
		return;
	}

	menu = console->menu;

	//thread the process and/or callback, or								 //zz debug
	//currently driven by GLUT key events processed within the main GLUT app loop

	//store callback function with console, called upon sucessful item selection
	console->pMenuCallback = pMenuCallback;

	//set the initial menu page
	console->page = 0;

	console->mode = kNPconsoleMenu;	//change to menu mode and pause messages

	//register console for user input events		
	//	npNewEventHandler (kNPeventKey, pConsoleKeyHandler, dataRef);		//zz debug
	//	npNewEventHandler (kNPeventMouse, pConsoleMouseHandler, dataRef);
	//	npNewEventHandler (kNPeventHID, pConsoleMouseHandler, dataRef);

	npUpdateConsoleMenu (console, data);
}

